/**
 * EidoKropki -- Web-based SGF Editor for Kropki
 * Copyright (c) 2008,2009, Bartek Dyda <kropki@yahoo.co.uk>
 *
 * Most of EidoKropki code comes from EidoGo, a Web-based SGF Go Editor,
 * Copyright (c) 2007, Justin Kramer <jkkramer@gmail.com>
 *
 * Code licensed under AGPLv3:
 * http://www.fsf.org/licensing/licenses/agpl-3.0.html
 *
 * Kropki board-related stuff
 */

/**
 * @class Keeps track of board state and passes off rendering to a renderer.
 * We can theoretically have any kind of renderer. The board state is
 * independent of its visual presentation.
 */
eidogo.Board = function() {
    this.init.apply(this, arguments);
};
eidogo.Board.prototype = {
  WHITE: 2,
  BLACK: 1,
    EMPTY: 0,
  INSTOP: 4,              // OR this to WHITE/BLACK/EMPTY to indicate that we're inside a stop
  COUNTED: 8,             // if (point & COUNTED), then this point inside a stop is already counted
  COUNTEDPUNISH: 16,      // if (point & COUNTEDPUNISH), then our point inside our stop is already counted as a punish
    /**
     * @constructor
     * @param {Object} The renderer to use to draw the board. Renderers must
     * have at least three methods: clear(), renderStone(), and renderMarker()
     * @param {Number} Board size -- theoretically could be any size,
     * but there's currently only CSS for 9, 13, and 19 and ...
     */
  init: function(renderer, boardSize, boardSizeY) {
        this.boardSize = boardSize || 19;
        this.boardSizeY = boardSizeY || 19;
        this.stones = this.makeBoardArray(this.EMPTY);
        this.markers = this.makeBoardArray(this.EMPTY);
	this.stops = "";
	this.territory = "";
	this.currStops = {};
	this.currStops.st = "";   // current, not-finished stops
	this.currStops.col = this.EMPTY;   // color of the current (not-finished) stops
        this.captures = {};
        this.captures.W = 0;
        this.captures.B = 0;
        this.captures.Wterr = 0;
        this.captures.Bterr = 0;
        this.cache = [];
	this.renderer = renderer || new eidogo.BoardRendererHtml();
        this.lastRender = {
            stones: this.makeBoardArray(null),
	    stops: "",
	    territory: "",
            markers: this.makeBoardArray(null)
        };
    },
    reset: function() {
        this.init(this.renderer, this.boardSize, this.boardSizeY);
    },
    clear: function() {
        this.clearStones();
        this.clearMarkers();
        this.clearCaptures();
	this.clearStopsAndTerr();
	this.clearCurrStops();
    },
    clearStones: function() {
        // we could use makeBoardArray(), but this is more efficient
        for (var i = 0; i < this.stones.length; i++) {
            this.stones[i] = this.EMPTY;
        }
    },
    clearMarkers: function() {
        for (var i = 0; i < this.markers.length; i++) {
            this.markers[i] = this.EMPTY;
        }
    },
    clearCaptures: function() {
        this.captures.W = 0;
        this.captures.B = 0;
        this.captures.Wterr = 0;
        this.captures.Bterr = 0;
    },
    clearStopsAndTerr: function() {
        this.stops = "";
	this.territory = "";
    },
    clearCurrStops: function() {
        this.currStops.st = "";
        this.currStops.col = this.EMPTY;
    },
    makeBoardArray: function(val) {
        // We could use a multi-dimensional array but doing this avoids
        // the need for deep copying during commit, which is very slow.
        return [].setLength(this.boardSize * this.boardSizeY, val);
    },
    /**
     * Save the current state. This allows us to revert back
     * to previous states for, say, navigating backwards in a game.
     */
    commit: function() {
        this.cache.push({
            stones: this.stones.concat(),
            stops: this.stops,
            territory: this.territory,
            currStops: {st: this.currStops.st, col: this.currStops.col},
	    captures: {W: this.captures.W, B: this.captures.B, 
		Wterr: this.captures.Wterr, Bterr: this.captures.Bterr}
        });
    },
    /**
     * Undo any uncomitted changes.
     */
    rollback: function() {
        if (this.cache.last()) {
            this.stones = this.cache.last().stones.concat();
	    this.stops  = this.cache.last().stops,
	    this.territory  = this.cache.last().territory,
	    this.currStops.st  = this.cache.last().currStops.st;
	    this.currStops.col = this.cache.last().currStops.col;
            this.captures.W = this.cache.last().captures.W;
            this.captures.B = this.cache.last().captures.B;
            this.captures.Wterr = this.cache.last().captures.Wterr;
            this.captures.Bterr = this.cache.last().captures.Bterr;
        } else {
            this.clear();
        }
    },
    /**
     * Revert to a previous state.
     */
    revert: function(steps) {
        steps = steps || 1;
        this.rollback();
        for (var i = 0; i < steps; i++) {
            this.cache.pop();
        }
        this.rollback();
    },
    addStone: function(pt, color) {
        this.stones[pt.y * this.boardSize + pt.x] = color;
    },

  wsp: function(i,j) {
    // change coordinates of ''2-dim array'' indexed by
    //  [-1,...,boardSize] x [-1,...,boardSizeY]
    // to 1-dim coordinates
    // We take [-1,...,boardSize] instead of [0,...,boardSize-1] to ease
    // checking if we're inside the board
        var tbsm1 = this.boardSizeY - 1;    // conversion to a number
        return (i+1)*(tbsm1+3)+j+1;
    },

  wsp_x: function(i) {
	// take the x-coordinate of the point-index of ''2-dim array'' [-1,...,boardSize] x [-1,...,boardSizeY]
	var tbsm1 = this.boardSizeY - 1;    // conversion to a number
	return Math.floor(i / (tbsm1+3)) - 1;
    },
  wsp_y: function(i) {
	// take the y-coordinate of the point-index of ''2-dim array'' [-1,...,boardSize] x [-1,...,boardSizeY]
	var tbsm1 = this.boardSizeY - 1;    // conversion to a number
	return (i % (tbsm1+3)) - 1;
    },

    sgfCoordToPoint: function(coord) {
        if (!coord) return {x: null, y: null};
        var sgfCoords = {
            a: 0, b: 1, c: 2, d: 3, e: 4, f: 5, g: 6, h: 7, i: 8, j: 9,
            k: 10,l: 11, m: 12, n: 13, o: 14, p: 15, q: 16, r: 17, s: 18, t:19,
	    u: 20, v: 21, w:22, x:23, y:24, z: 25,
	    A: 26, B: 27, C: 28, D: 29, E: 30, F: 31, G: 32, H: 33, I: 34, J: 35,
	    K: 36, L: 37, M: 38, N: 39 
        };
        return {
            x: sgfCoords[coord.charAt(0)],
            y: sgfCoords[coord.charAt(1)]
        };
    },

    pointToSgfCoord: function(pt) {
        if (!pt || pt.x<0 || pt.x>=this.boardSize || pt.y<0 || pt.y>=this.boardSizeY)
            return null;
         var pts = {
            0: 'a', 1: 'b', 2: 'c', 3: 'd', 4: 'e', 5: 'f', 6: 'g', 7: 'h',
            8: 'i', 9: 'j', 10: 'k', 11: 'l', 12: 'm', 13: 'n', 14: 'o',
            15: 'p', 16: 'q', 17: 'r', 18: 's', 19: 't', 20: 'u', 21: 'v',
	    22: 'w', 23: 'x', 24: 'y', 25: 'z',
	    26: 'A', 27: 'B', 28: 'C', 29: 'D', 30: 'E', 31: 'F', 32: 'G',
	    33: 'H', 34: 'I', 35: 'J', 36: 'K', 37: 'L', 38: 'M', 39: 'N'
        };
        return pts[pt.x] + pts[pt.y];
    },
  floodFill: function(tab) {
	// tab: array of (x+2) x (y+2) size, with forbidden points marked by 4
	//  (forbidden may be: border of a stop - if it is given, or points
	//   of one of the players, if automatic stops are to be found),
	// and the rest is marked by 0
    var na_stosie = 0;
    var stos = new Array();
    var imax = this.boardSize-1;
    var jmax = this.boardSizeY-1;
    // mark upper and lower margin by 4 and push on the stack 'stos',
    // if it's not forbidden; also mark by 4 the margins
    // outside the board
    for (var i=-1; i<=imax+1; i++) {
      tab[this.wsp(i,-1)]=4;   tab[this.wsp(i,jmax+1)]=4;
      if (i>=0 && i<=imax) {
	var pole = this.wsp(i,0);
	if (tab[pole]==0) {  // not forbidden
	  tab[pole]=4;
	  stos[na_stosie++] = pole;
	}
	pole = this.wsp(i,jmax);
	if (tab[pole]==0) {  // not forbidden
	  tab[pole]=4;
	  stos[na_stosie++] = pole;
	}
      }
      else {
	tab[this.wsp(i,0)]=4;   tab[this.wsp(i,jmax)]=4;   // we're on the left/right margin outside
      }
    }
    // and the left and right sides remain
    for (var j=1; j<jmax; j++) {
      tab[this.wsp(-1,j)]=4;   tab[this.wsp(imax+1,j)]=4;
      var pole = this.wsp(0,j);
      if (tab[pole]==0) {  // not forbidden
	tab[pole]=4;
	stos[na_stosie++] = pole;
      }
      pole = this.wsp(imax,j);
      if (tab[pole]==0) {  // not forbidden
	tab[pole]=4;
	stos[na_stosie++] = pole;
      }
    }  

    // now visit the neighbours
    var dp = [-1, 1, jmax+3, -jmax-3];    // +-1 shifts horizontal/vertical
    while (na_stosie) {
      var pole = stos[--na_stosie];
      for (var k=0; k<4; k++)
	if (tab[pole+dp[k]]==0) {
	  tab[pole+dp[k]]=4;
	  stos[na_stosie++] = pole+dp[k];
	}
    }
    },

  markInterior: function(tab, pt) {
	// tab: array of (x+2) x (y+2) size, with enclosing player's dots marked by 4
	// and the rest is marked by 0;
	//  pt: point to enclose (it should be possible to enclose it)
	// output: interior points marked by 1, (potential) exterior changed from 4 to 5
	// returns index of the leftmost point of exterior
    var na_stosie = 0;
    var stos = new Array();
    var imax = this.boardSize-1;
    var jmax = this.boardSizeY-1;
    var leftmost = this.wsp(pt.x, pt.y);
    stos[na_stosie++] = this.wsp(pt.x, pt.y);

    var dp = [-1, 1, jmax+3, -jmax-3];    // +-1 shifts horizontal/vertical
    while (na_stosie) {

    do {
	var pole = stos[--na_stosie];
	// visit neighbours
	for (var k=0; k<4; k++)
	    if (tab[pole+dp[k]]==0) {
		tab[pole+dp[k]] = 1;
		stos[na_stosie++] = pole+dp[k];
	    }
	    else if (tab[pole+dp[k]]==4) {
		if (leftmost > pole+dp[k])
		    leftmost = pole+dp[k];
		tab[pole+dp[k]] = 5;
	    }
    }
    while (na_stosie);
    return leftmost;
    }
    },

  findStopBorder: function(tab, pt) {
	// tab: array of (x+2) x (y+2) size, with enclosing player's dots marked by 4,
	// exterior=5, interior=1, and the rest is marked by 0;
	//  pt: the leftmost upper point of the border (of least index)
	// output: dad table, with previous border point or 0, if not on the border
	var imax = this.boardSize-1;    // size = imax+1, eg. for 20x20 board imax=19
	var jmax = this.boardSizeY-1;
    	dad = new Array((imax+3)*(jmax+3));
	for (var i=0; i<(imax+3)*(jmax+3); i++)
	    dad[i] = 0;
	var kierunki_dx = [0,1,1,1, 0,-1,-1,-1];
	var kierunki_dy = [-1,-1,0,1, 1,1,0,-1];
	var kierunki_dind = new Array(8);
	for (var i=0; i<8; i++)
	    kierunki_dind[i]= kierunki_dx[i]*(jmax+3) + kierunki_dy[i];
	var kierunki = new Array();
	var points = new Array();
	var stack_pointer = 1;
	var leftmost = this.wsp(pt.x, pt.y);
	kierunki[0] = 0x74; // 7=kier_ost, 4=kier;  kiedys bylo 5 * 0x11
	dad[ points[0] = this.wsp(pt.x+1, pt.y-1) ] = leftmost;
	var next_point = false;   // simulate 'goto ...'
	do {
	    // Nastepny_punkt:
	    var  pn = points[stack_pointer-1];
	    var kier     = (kierunki[stack_pointer-1] & 0xf);
	    var kier_ost = (kierunki[stack_pointer-1] >> 4);
	    if (--kier==-1) kier=7;
	    while (kier!=kier_ost) {
		var pn2 = pn + kierunki_dind[kier];
		if (tab[pn2]==5 && dad[pn2]==0) {
		    dad[pn2] = pn;
		    if (pn2==leftmost)
			return dad;
		    // zapamietaj aktualny kierunek
		    kierunki[stack_pointer-1] &= 0xf0;
		    kierunki[stack_pointer-1] |= kier;
		    // wsadz na stos punkt pn2
		    points[stack_pointer] = pn2;
		    var kk=(kier^4);    // wez przeciwny kierunek do ,,kier''
		    kierunki[stack_pointer++] = (kk<<4) + kk;
		    next_point = true;  break;  // goto Nastepny_punkt;
		}
		if (--kier==-1) kier=7;
	    }
	    if (!next_point) {
		// przeszlismy cala galaz, zdejmij punkt ze stosu
		stack_pointer--;
	    }
	    else next_point = false;
	}
	while (stack_pointer);
	return null;  // no stop border...
    },

  makePointInside: function(pt, stopColor, newMask, in_terr, punish){
	  var stone      = this.getStone(pt);
	  var stoneColor = this.getStone(pt) & ~(this.INSTOP | this.COUNTED | this.COUNTEDPUNISH);
	  if (stoneColor == this.EMPTY) {
	    stone |= newMask;
	    if (in_terr) {
	      if (stopColor == this.BLACK)
		this.captures.Bterr++;
	      else
		this.captures.Wterr++;
	    }
	  }
	  else if (stone & this.INSTOP) {
	    // a stone was already in stop
	    if (stoneColor == stopColor && (stone & this.COUNTED)) {
		// we free our dot
		stone &= ~(this.COUNTED);
		if (stoneColor == this.BLACK)
		  this.captures.W--;
		else
		  this.captures.B--;
	    }
	    else if (stoneColor != stopColor && !(stone & this.COUNTED)) {
	      stone |= this.COUNTED;
	      // we capture enemy dot (which was in enemy's stop)
	      if (stopColor == this.BLACK)
		this.captures.B++;
	      else
		this.captures.W++;
	      }
	  }
	  else { // here we have a stone still not in stop
	    if (stoneColor == stopColor) {
	      stone |= newMask;   // we capture our dot
	      if (punish) {
		if (stopColor == this.BLACK)
		  this.captures.B--;
		else
		  this.captures.W--;
		stone |= this.COUNTEDPUNISH;
	      }
	    }
	    else {
	      if (stopColor == this.BLACK)
		this.captures.B++;
	      else
		this.captures.W++;
	      stone |= (newMask | this.COUNTED);   // we capture enemy dot
	    }
	  }
	  this.addStone(pt, stone);
    },

  makeStop: function(stop, in_terr, punish){
    //  if in_terr==0:
    // marks fields inside stop | this.INSTOP
    // count and mark killed dots
    //  if in_terr!=0:
    // count and mark killed dots, count territory
    // the border of a stop is given in an argument array
    // punish == true if there's punish of -1 point for surrounding own dots
    var maxx = (this.boardSize-1) + 1;   // a hack in case this.boardSize is a string
    var maxy = (this.boardSizeY-1) + 1;
    
    var dad = new Array((maxx+2)*(maxy+2));
    var dad2 = new Array((maxx+2)*(maxy+2));
    for (var i=0; i<(maxx+2)*(maxy+2); i++) {
      dad[i] = 0;
      dad2[i] = 0;
    }

    var pocz_pole=-1;
    var stopColor = this.EMPTY;
    var poprz_pole = -1;
    for(var n=0; n<stop.length; n+=2){
      var pt = this.sgfCoordToPoint(stop.substr(n, 2));
      var pole = this.wsp(pt.x, pt.y);
      if (stopColor == this.EMPTY)
	stopColor = this.getStone(pt) & ~(this.INSTOP | this.COUNTED | this.COUNTEDPUNISH);
      if (poprz_pole>0) {
	  dad[pole] = poprz_pole;   // zaznacz brzeg stopu
	  dad2[poprz_pole] = pole;
      }
      poprz_pole = pole;
      if (pocz_pole==-1) {
	pocz_pole = pole;
      }
      else if (pocz_pole == pole)
	break;   // the line segment is closed now
    }


    var newMask = (in_terr == 0) ? this.INSTOP : 0;
    for (var i=0; i<maxx; i++) {
	var przeciec = 0;
	var zktorej = 0;
	var ind = this.wsp(i, 0);
	for (var j=0; j<maxy; j++) {
	    var pt = {x: i,  y: j};
	    if (dad[ind]) {
		var d = (this.wsp_x(dad[ind]) - i) + (this.wsp_x(dad2[ind]) - i);
		if (zktorej==0) {   // przed chwila nie bylismy na brzegu
		    if (d==0) {
			przeciec ^= 1;   // przecinamy
		    }
		    else if (d==1 || d==-1) {
			zktorej = d;
		    }
		}
		else {  // przed chwila BYLISMY na brzegu
		    if (d) {   // gdy d==0, to dalej bedziemy na brzegu, wiec tylko przypadek d!=0 jest ciekawy
			if (d!=zktorej)
			    przeciec ^= 1;  // przecinamy, a nie przechodzimy bokiem
			zktorej = 0;   // wychodzimy z brzegu
		    }
		}
	    }
	    else if (przeciec) {
		// pole (ind) = (pt) jest wewnatrz stopu!
		this.makePointInside(pt, stopColor, newMask, in_terr, punish);
	    }
	    ind++;
	}
	//assert(przeciec==0);
    }

    },

  addStop: function(coord, punish) {
      var stopy = coord.split(".");
      for (var n=stopy.length-1; n>=0; n--) {
	this.makeStop(stopy[n], 0, punish);
      }
      this.stops += coord;
    },

  addTerritory: function(coord, punish) {
      var terr = coord.split(".");
      for (var n=terr.length-1; n>=0; n--) {
	this.makeStop(terr[n], 1, punish);
      }
      this.territory += coord;
    },


    addCurrStop: function(x, y, color) {
      // check whether its legal to add (x,y) to 'curr', eg. curr="ab."
      // x,y  = next point (clicked)
      // color = "B" or "W"
      // returns curr+(x+y)[.] if (x,y) is a valid next point, and "" otherwise
      this.currStops.col = (color == "B") ? this.BLACK : this.WHITE;
      var coord = this.pointToSgfCoord({x: x, y: y});
      if (this.currStops.st=="") {
	if (this.getStone({x: x, y: y}) != this.EMPTY)
	  return this.currStops.st;
	else
	  return this.currStops.st += coord + ".";
      }
      if (coord!=this.currStops.st.substr(0,2) && this.getStone({x: x, y: y}) != this.currStops.col)
	return this.currStops.st;
      if (this.currStops.st.lastIndexOf(".") < this.currStops.st.length-1) {
	var last = this.sgfCoordToPoint(this.currStops.st.substr(this.currStops.st.length-2,2));
	if (x < last.x-1 || x > last.x+1 || y < last.y-1 || y > last.y+1)
	  return this.currStops.st;   // new point to far from the last
	if (x==last.x && y==last.y)
	  return (this.currStops.st = this.currStops.st.substr(0,this.currStops.st.length-2));
	if (this.currStops.st.substr( this.currStops.st.lastIndexOf(".")+1, 2) == coord)
	  return this.currStops.st += coord + ".";
	else return this.currStops.st += coord;
      }
      else {
	// this is the first point of the stop
	return this.currStops.st += coord;
      }
    },
    finishCurrStop: function() {
      // user pressed 'stop' for the second time, prepare the current stop
      if (this.currStops.st=="") return;
      this.currStops.st = this.currStops.st.substr(0, this.currStops.st.lastIndexOf("."));
      if (this.currStops.st.length <= 2) return;
      var stops = this.currStops.st.substr(3, this.currStops.st.length-3).split(".");
      for (var i=0; i<stops.length; i++) {
	// split stops[i] into pairs of letters, don't take the last pair (which is the same as the first)
	var pairs = new Array(stops[i].length/2 - 1);
	var min=0;    // index of a minimal pair
	for (var p=0; p<stops[i].length-2; p+=2) {
	  pairs[p/2] = stops[i].substr(p,2);
	  if (pairs[p/2] < pairs[min])
	    min = p/2;
	}
	var prevp = ((min>0) ? pairs[min-1] : pairs[pairs.length-1]);
	var nextp = ((min<pairs.length-1) ? pairs[min+1] : pairs[0]);
	var step;
	if (prevp.charAt(0)==pairs[min].charAt(0))
	  step = -1;
	else if (nextp.charAt(0)==pairs[min].charAt(0))
	  step = +1;
	else
	  step = (nextp > prevp) ? +1 : -1;
	// now glue the stop again
	stops[i] = pairs[min];
	for (var p=min+step; ; p+=step) {
	  if (p==pairs.length) p=0;
	  else if (p==-1) p=pairs.length-1;
	  stops[i] += pairs[p];
	  if (p==min) break;
	}
      }
      stops.sort();
      this.currStops.st = this.currStops.st.substr(0, 2);
      for (var i=0; i<stops.length; i++)
	this.currStops.st += "." + stops[i];
    },
    doAutoStops: function(colour) {
	// for Russian rules
	var imax = this.boardSize-1;    // size = imax+1, eg. for 20x20 board imax=19
	var jmax = this.boardSizeY-1;
    
	var tab = new Array((imax+3)*(jmax+3));
	for (var i=0; i<(imax+3)*(jmax+3); i++)
	    tab[i] = 0;

	// now mark all our (=colour) dots by 4 and then floodfill the exterior...
	for (var i=0; i<=imax; i++)
	    for (var j=0; j<=jmax; j++) {
		var pt = {x: i,  y: j};
		var stone      = this.getStone(pt);
		var stoneColor = stone & ~(this.INSTOP | this.COUNTED | this.COUNTEDPUNISH);
		if (stoneColor == colour && !(stone & this.INSTOP))
		    tab[this.wsp(i,j)] = 4;   // mark our dot
	    }
	this.floodFill(tab);
	// enemy dots that are not marked by 4 should be enclosed
	var oppColor = (colour == this.WHITE ? this.BLACK : this.WHITE);
	for (var i=1; i<imax; i++)
	    for (var j=1; j<jmax; j++) {
		var pt = {x: i,  y: j};
		if (tab[this.wsp(i,j)]==0) {
		    var stone      = this.getStone(pt);
		    var stoneColor = this.getStone(pt) & ~(this.INSTOP | this.COUNTED | this.COUNTEDPUNISH);
		    if (stoneColor == oppColor && !(stone & this.INSTOP)) {
			// capture this stone...
			var leftmost = this.markInterior(tab, pt);
			var dad = this.findStopBorder(tab, {x: this.wsp_x(leftmost), y: this.wsp_y(leftmost)} );
			// save the border in the sgf-style string
			var stop = ".";
			var p=leftmost;
			do {
			    stop += this.pointToSgfCoord({x: this.wsp_x(p), y: this.wsp_y(p)});
			    p = dad[p];
			}
			while (p != leftmost);
			stop += this.pointToSgfCoord({x: this.wsp_x(p), y: this.wsp_y(p)});
			this.addStop(stop, false);
			// correct the marks to find another stops...
			for (var ii=0; ii<=imax; ii++)
			    for (var jj=0; jj<=jmax; jj++) {
				var pt = {x: ii,  y: jj};
				var stone      = this.getStone(pt);
				var stoneColor = stone & ~(this.INSTOP | this.COUNTED | this.COUNTEDPUNISH);
				if (stoneColor == colour && !(stone & this.INSTOP))
				    tab[this.wsp(ii,jj)] = 4;   // mark our dot
				else if (tab[this.wsp(ii,jj)] == 1 || tab[this.wsp(ii,jj)] == 5)
				    tab[this.wsp(ii,jj)] = 0;   // our dot is now inside stop...
			    }
		    }
		}
	    }
    },
  
    whoMakesLastMove: function(whosmove) {
      var emptyf = 0;
      for (var i=0; i<this.boardSize; i++)
	for (var j=0; j<this.boardSizeY; j++) {
	  var pt = {x: i,  y: j};
	  var stone      = this.getStone(pt);
	  if (stone == this.EMPTY)   // empty place not in stop
	    emptyf++;
	}
      if ((emptyf % 2) == 1)
	return whosmove;
      else return (whosmove == this.BLACK ? this.WHITE : this.BLACK);
    },
    getGameResult: function(rules,  whoslast) {
	var surplus = this.captures.Wterr - this.captures.Bterr;
	var has_more_terr = this.WHITE;
	if (surplus < 0) {
	  surplus = -surplus;
	  has_more_terr = this.BLACK;
	}
	switch (rules) {
	case 0:
	// standard szkrab rules
	  if (whoslast != has_more_terr && surplus>0)
	    surplus--;
	  if ((surplus % 2) == 1)  // round down
	    surplus--;
	  var resB = this.captures.B;
	  var resW = this.captures.W;
	  if (has_more_terr == this.BLACK)
	    resB += surplus/2;
	  else
	    resW += surplus/2;
	  return {W: resW, B: resB};
	case 1:
	  // pass-move-allowed rules
	  return {W: (this.captures.W + ((has_more_terr==this.BLACK) ? 0 : (surplus/2))),
		  B: (this.captures.B + ((has_more_terr==this.BLACK) ? (surplus/2) : 0))};
	case 2:
	// lastdotsafe=0
	  if ((surplus % 2) == 1) {  // round
	      if (whoslast == has_more_terr)
		  surplus--;
	      else
		  surplus++;
	  }
	  var resB = this.captures.B;
	  var resW = this.captures.W;
	  if (has_more_terr == this.BLACK)
	    resB += surplus/2;
	  else
	    resW += surplus/2;
	  return {W: resW, B: resB};
	}
    },
    getCurrStop: function() {
      return this.currStops.st;
    },
    isOnBoard: function(pt) {
	return (pt.x>=0 && pt.x<this.boardSize && pt.y>=0 && pt.y<this.boardSizeY);
    },
    getStone: function(pt) {
    /*
       var p = this.stones[pt.y * this.boardSize + pt.x];
       return p & ~(this.INSTOP | THIS.COUNTED | this.COUNTEDPUNISH);
    */
       return this.stones[pt.y * this.boardSize + pt.x];
    },
    isInStop: function(pt) {
        return (this.stones[pt.y * this.boardSize + pt.x] & this.INSTOP) != 0;
    },
    getRegion: function(t, l, w, h) {
        var region = [].setLength(w * h, this.EMPTY);
        var offset;
        for (var y = t; y < t + h; y++) {
            for (var x = l; x < l + w; x++) {
                offset = (y - t) * w + (x - l);
                region[offset] = this.getStone({x:x, y:y});
            }
        }
        return region;
    },
    addMarker: function(pt, type) {
        this.markers[pt.y * this.boardSize + pt.x] = type;
    },
    getMarker: function(pt) {
        return this.markers[pt.y * this.boardSize + pt.x];
    },
    render: function(complete) {
        var stones = this.makeBoardArray(null);
        var markers = this.makeBoardArray(null);
	var stops;
        var color, type;
        var len;
        if (!complete && this.cache.last()) {
            var lastCache = this.cache.last();
            len = this.stones.length;
            // render only points that have changed since the last render
            for (var i = 0; i < len; i++) {
                if (lastCache.stones[i] != this.lastRender.stones[i]) {
                    stones[i] = lastCache.stones[i];
                }
            }
            markers = this.markers;
	    if (lastCache.stops.indexOf(this.lastRender.stops) == 0) {
	      // this doesn't work well -- doesn't clear stops taken back
	      // stops = lastCache.stops.substr(this.lastRender.stops.length);
	      stops = this.stops;
	      this.lastRender.stops = "";
	      this.renderer.renderStop0();
	    }
	    else {
	      stops = this.stops;
	      this.lastRender.stops = "";
	      this.renderer.renderStop0();
	    }
        } else {
            // render everything
            stones = this.stones;
            markers = this.markers;
	    stops = this.stops;
	    this.lastRender.stops = "";
	    this.renderer.renderStop0();
        }
        var offset;

	this.renderer.adjustSizes();

        for (var x = 0; x < this.boardSize; x++) {
            for (var y = 0; y < this.boardSizeY; y++) {
                offset = y * this.boardSize + x;
                if (markers[offset] != null) {
                    this.renderer.renderMarker({x: x, y: y}, markers[offset]);
                    this.lastRender.markers[offset] = markers[offset];
                }
                if (stones[offset] == null) {
                    continue;
                } else {
		  var stone_ = stones[offset] & ~(this.INSTOP | this.COUNTED | this.COUNTEDPUNISH);
		  if (stone_ == this.EMPTY) {
                    color = "empty";
		  } else {
                    color = (stone_ == this.WHITE ? "white" : "black");
		  }
		}
                this.renderer.renderStone({x: x, y: y}, color);
                this.lastRender.stones[offset] = stones[offset];
            }
        }


	this.renderer.renderCurrStop0();
	this.renderer.renderBackground();

	var tabStops = new Array(stops, this.currStops.st,  this.territory);
	if (this.currStops.st!="")
	  tabStops[1] = this.currStops.st.substr(2,this.currStops.st.length-2);
	for (var cs=0; cs<3; cs++)
	  if (tabStops[cs]!="") {
	    if (cs==0) this.lastRender.stops += stops;
	    var stopy = tabStops[cs].split(".");
	    if (stopy.length) {
	      for (var s = stopy.length-1; s>=0; s--) {
		var ptx = new Array();
		var pty = new Array();
		var i=0;
		var color = this.EMPTY;
		for(var n=0; n<stopy[s].length; n+=2){
		  var pt =  this.sgfCoordToPoint( stopy[s].substr(n,2));
		  if (color==this.EMPTY) {
		      color=this.getStone(pt) & ~(this.INSTOP | this.COUNTED | this.COUNTEDPUNISH);
		  }
		  ptx[i] = pt.x;  pty[i] = pt.y;
		  i++;
		}
		if (color==this.WHITE) color = "white";
		else if (color==this.BLACK) color = "black";
		else color = "empty";
		this.renderer.renderStop( ptx, pty, color, cs);
	      }
	    }
	  }
	// render the dot in the current stop
	if (this.currStops.st!="") {
	  var pt = this.sgfCoordToPoint(this.currStops.st.substr(0,2));
	  if ((this.currStops.col & ~(this.INSTOP | this.COUNTED | this.COUNTEDPUNISH)) ==this.EMPTY)
	    color = "empty";
	  else
	    color = (this.currStops.col==this.WHITE) ? "white" : "black";
	  this.renderer.renderStone(pt, color);
	}
	this.renderer.renderStop1();
	this.renderer.renderCurrStop1();
  }
};

/**
 * @class An HTML/DOM-based board renderer.
 */
eidogo.BoardRendererHtml = function() {
    this.init.apply(this, arguments);
}
eidogo.BoardRendererHtml.prototype = {
    /**
     * @constructor
     * @param {HTMLElement} domContainer Where to put the board
     */
    init: function(domContainer, boardSize, boardSizeY, player, crop) {
        if (!domContainer) {
            throw "No DOM container";
            return;
        }
        this.boardSize = boardSize || 19;
        this.boardSizeY = boardSizeY || 19;
        var domGutter = document.createElement('div');
        domGutter.className = "board-gutter";
	if (this.boardSize == 20 && this.boardSize==this.boardSizeY)
	  domGutter.className += " with-coords20";
	else if (this.boardSize == 30 && this.boardSize==this.boardSizeY)
	  domGutter.className += " with-coords30";
        domContainer.appendChild(domGutter);
        var domBoard = document.createElement('div');
	//	if ((this.boardSize==9 || this.boardSize==13 || this.boardSize==19 || this.boardSize==20 || 
	if ((this.boardSize==20 || this.boardSize==30) && (this.boardSize==this.boardSizeY))
	    domBoard.className = "board size" + this.boardSize;
	else
	    domBoard.className = "board sizeany";
        domBoard.style.position = (crop && eidogo.browser.ie ? "static" : "relative");
        domGutter.appendChild(domBoard);
        this.domNode = domBoard;
        this.domGutter = domGutter;
        this.domContainer = domContainer;
        this.player = player;
        this.uniq = domContainer.id + "-";
	var maxXY = (this.boardSize >= this.boardSizeY) ? this.boardSize : this.boardSizeY;
        this.renderCache = {
            stones: [].setLength(maxXY, 0).addDimension(maxXY, 0),
            markers: [].setLength(maxXY, 0).addDimension(maxXY, 0)
        }
        // auto-detect point width, point height, and margin
        this.pointWidth = 0;
        this.pointHeight = 0;
        this.margin = 0;
        var stone = this.renderStonePng({x:0,y:0}, "black");
	this.pointWidthSave = this.pointWidth = this.pointHeight = stone.offsetWidth;
        this.renderStonePng({x:0,y:0}, "white"); // just for image caching
        this.renderMarker({x:0,y:0}, "current"); // just for image caching
        this.clear();
	this.domNodeOffsetWidthSave = this.domNode.offsetWidth;
        this.margin = (this.domNode.offsetWidth - (this.boardSize * this.pointWidth)) / 2;
        
        // needed to accommodate IE's broken layout engine
        this.scrollX = 0;
        this.scrollY = 0;
        
        if (crop) {
            this.crop(crop);
            if (eidogo.browser.ie) {
                var parent = this.domNode.parentNode;
                while (parent && parent.tagName && !/^body|html$/i.test(parent.tagName)) {
                    this.scrollX += parent.scrollLeft;
                    this.scrollY += parent.scrollTop;
                    parent = parent.parentNode;
                }
            }
        }
        
	this.scrollXsave = this.scrollX;
	this.scrollYsave = this.scrollY;

        // add the search region selection box for later use
        this.dom = {};
        this.dom.searchRegion = document.createElement('div');
        this.dom.searchRegion.id = this.uniq + "search-region";
        this.dom.searchRegion.className = "search-region";
        this.domNode.appendChild(this.dom.searchRegion);
        
        eidogo.util.addEvent(this.domNode, "mousemove", this.handleHover, this, true);
        eidogo.util.addEvent(this.domNode, "mousedown", this.handleMouseDown, this, true);
        eidogo.util.addEvent(this.domNode, "mouseup", this.handleMouseUp, this, true);

	// add div's for drawing finished stops and for drawing current stops
	this.dom.stops = document.createElement('div');
        this.dom.stops.id = this.uniq + "stops";
        this.dom.stops.className = "stops";
        this.domNode.appendChild(this.dom.stops);

	this.jra = Raphael(domBoard, 190, 190); //430, 430);
	this.jra_arr = [];
	this.jra_arr_cs = [];  // current stops here

	this.stones_el = [].setLength(maxXY, 0).addDimension(maxXY, 0);
	for (var i=0; i<=this.boardSize-1; i++)
	    for (var j=0; j<=this.boardSizeY-1; j++) {
		this.stones_el[i][j] = null;
	    }

	this.dom.currStops = document.createElement('div');
        this.dom.currStops.id = this.uniq + "currStops";
        this.dom.currStops.className = "currStops";
        this.domNode.appendChild(this.dom.currStops);

    },
    adjustSizes: function() {
      var maxXY = (this.boardSize >= this.boardSizeY) ? this.boardSize : this.boardSizeY;
      if (maxXY <= 20) {
	this.pointWidth = this.pointHeight =this.pointWidthSave;
	this.scrollX = this.scrollXsave;
	this.scrollY = this.scrollYsave;
      }
      else if (maxXY == 30) {
	this.pointWidth = this.pointHeight = 13;
	this.scrollX = this.scrollXsave;
	this.scrollY = this.scrollYsave;
      }
      else if (maxXY >=36 && maxXY <= 40) {
	this.pointWidth = this.pointHeight = 10;
	this.scrollX = this.scrollXsave;
	this.scrollY = this.scrollYsave;
      }
      else {
	  this.pointWidth = this.pointHeight = 
	  Math.floor((this.pointWidthSave * (20-1)) / (maxXY-1));
	this.scrollX = this.scrollXsave;
	this.scrollY = this.scrollYsave;
      }
      this.margin = ((this.domNodeOffsetWidthSave - (maxXY * this.pointWidth)) / 2);
    },
    showRegion: function(bounds) {
        this.dom.searchRegion.style.top = (this.margin + this.pointHeight * bounds[0]) + "px";
        this.dom.searchRegion.style.left = (this.margin + this.pointWidth * bounds[1]) + "px";
        this.dom.searchRegion.style.width = this.pointWidth * bounds[2] + "px";
        this.dom.searchRegion.style.height = this.pointHeight * bounds[3] + "px";
        eidogo.util.show(this.dom.searchRegion);
    },
    hideRegion: function() {
        eidogo.util.hide(this.dom.searchRegion);  
    },
    clear: function() {
        this.domNode.innerHTML = "";
    },
    renderStone: function(pt, color) {
	var point_size = 0.25 * this.pointWidth;
	if (this.pointWidth<=10)
	    point_size = 0.35 * this.pointWidth;
	if (color!="empty") {
	    if (!this.stones_el[pt.x][pt.y])
		this.stones_el[pt.x][pt.y] =
		    this.jra.circle(pt.x *  this.pointWidth+ this.pointWidth/2 + this.margin - this.scrollX,
				    pt.y *  this.pointWidth+ this.pointWidth/2 + this.margin - this.scrollY,
				    point_size);
	    else
		this.stones_el[pt.x][pt.y].attr(
			{cx: pt.x *  this.pointWidth+ this.pointWidth/2 + this.margin - this.scrollX,
				cy: pt.y *  this.pointWidth+ this.pointWidth/2 + this.margin - this.scrollY,
				r: point_size});
	    if (color=="black")
		this.stones_el[pt.x][pt.y].attr({fill: "#1f33c5", stroke: "#1f33c5"}).show();
	    else
		this.stones_el[pt.x][pt.y].attr({fill: "#e3241a", stroke: "#e3241a"}).show();
	    this.stones_el[pt.x][pt.y].toFront();
	}
	else {
	    if (this.stones_el[pt.x][pt.y]) {
		this.stones_el[pt.x][pt.y].remove();
		this.stones_el[pt.x][pt.y] = null;
	    }
	}
    },
    renderStonePng: function(pt, color) {
        var stone = document.getElementById(this.uniq + "stone-" + pt.x + "-" + pt.y);
        if (stone) {
            stone.parentNode.removeChild(stone);
        }
        if (color != "empty") {
            var div = document.createElement("div");
            div.id = this.uniq + "stone-" + pt.x + "-" + pt.y;
	    var maxXY = (this.boardSize >= this.boardSizeY) ? this.boardSize : this.boardSizeY;
	    if (maxXY <= 20)
	      div.className = "point stone " + color;
	    else if (maxXY < 35)
	      div.className = "point stone small " + color;
	    else
	      div.className = "point stone tiny " + color;
            try {
                div.style.left = (pt.x * this.pointWidth + this.margin - this.scrollX) + "px";
                div.style.top = (pt.y * this.pointHeight + this.margin - this.scrollY) + "px";
            } catch (e) {}
            this.domNode.appendChild(div);
            return div;
        }
        return null;
    },
    renderMarker: function(pt, type) {
        if (this.renderCache.markers[pt.x][pt.y]) {
            var marker = document.getElementById(this.uniq + "marker-" + pt.x + "-" + pt.y);
            if (marker) {
                marker.parentNode.removeChild(marker);
            }
        }
        if (type == "empty" || !type) { 
            this.renderCache.markers[pt.x][pt.y] = 0;
            return null;
        }
        this.renderCache.markers[pt.x][pt.y] = 1;
        if (type) {
            var text = "";
            switch (type) {
                case "triangle":
                case "square":
                case "circle":
                case "ex":
                case "territory-white":
                case "territory-black":
                case "dim":
                case "current":
                    break;
                default:
                    if (type.indexOf("varB:") == 0) {
                        text = type.substring(5);
			if (this.pointWidth>=13)
			    type = "variationB";
			else
			    type = "variationBsmall";
                    } else if (type.indexOf("varW:") == 0) {
                        text = type.substring(5);
			if (this.pointWidth>=13)
			    type = "variationW";
			else
			    type = "variationWsmall";
                     } else {
                        text = type;
                        type = "label";
                    }
                    break;
            }
            var div = document.createElement("div");
            div.id = this.uniq + "marker-" + pt.x + "-" + pt.y;
            div.className = "point marker " + type;
	    var correction = 0;
	    var maxXY = (this.boardSize >= this.boardSizeY) ? this.boardSize : this.boardSizeY;
	    if (maxXY>=35)
	      correction = 4;
	    else if (maxXY>27)
	      correction = 3;
	    else if (maxXY>20)
	      correction = 2;
            try {
                div.style.left = (pt.x * this.pointWidth + this.margin - correction - this.scrollX) + "px";
                div.style.top = (pt.y * this.pointHeight + this.margin - correction - this.scrollY) + "px";
            } catch (e) {}
            div.appendChild(document.createTextNode(text));
            this.domNode.appendChild(div);
            return div;
        }
        return null;
    },


    renderStop: function(ptx, pty, color, type) {
    if (ptx.length < 2) return;
    for (var i=0; i<ptx.length; i++) {
      ptx[i] = ptx[i] * this.pointWidth+ this.pointWidth/2 + this.margin - this.scrollX;
      pty[i] = pty[i] * this.pointWidth+ this.pointWidth/2 + this.margin - this.scrollY;
    }
    var path = "M "+ptx[0] + ' ' + pty[0] + ' ';
    for (var i=1; i<ptx.length; i++) {
	path += "L " + ptx[i] + ' ' + pty[i];
    }

    var thisstop;

    switch (type) {
      case 0: {
        if (color == "white")
	    thisstop = this.jra.path(path).attr({fill: "#eaa297", stroke: "black"});
	else if (color == "black")
	    thisstop = this.jra.path(path).attr({fill: "#b0b7f4", stroke: "black"});
	else
	    thisstop = this.jra.path(path).attr({fill: "red", stroke: "black"});
	this.jra_arr.push(thisstop);
	break;
      }
      case 2: {
        if (color == "white")
	    thisstop = this.jra.path(path).attr({fill: "#ffb2a6", stroke: "#ffb2a6"});
	else if (color == "black")
	    thisstop = this.jra.path(path).attr({fill: "#c2c9ff", stroke: "#c2c9ff"});
	else
	    thisstop = this.jra.path(path).attr({fill: "red", stroke: "red"});
	this.jra_arr.push(thisstop);
	break;
      }
      case 1: {
	  thisstop = this.jra.path(path).attr({stroke: "#00ff00", "stroke-width": 3 });
	  this.jra_arr_cs.push(thisstop);   // cs = curr stop!
	  break;
      }
    }
    thisstop.toBack();
    },

    renderStop0: function() {
    // clear all stops
      while (this.jra_arr.length > 0) {
	  var el = this.jra_arr.pop();
	  el.remove();
      }
    },
    renderStop1: function() {
    // paint all stops
	//      this.jg.paint();
    },
    renderBackground: function() {
    // renders background for non-standard boards
    var maxXY = (this.boardSize >= this.boardSizeY) ? this.boardSize : this.boardSizeY;
    var stand_size = (maxXY==20 || maxXY==30);
    if (!stand_size || this.boardSize != this.boardSizeY) {
      var min = (this.pointWidth/2 + this.margin);
      var max = ( (this.boardSize-1) *  this.pointWidth+ this.pointWidth/2 + this.margin );
      var maxY = ( (this.boardSizeY-1) *  this.pointWidth+ this.pointWidth/2 + this.margin);
      // render lines 1,2,3,4, 6,7,8,9, 11,12,...
      var path = "";
      for (var i=0; i<maxXY; i++) {
	  if (i % 5==0) continue;
	  if (i<this.boardSize) {
	      var x = Math.floor(i *  this.pointWidth+ this.pointWidth/2 + this.margin - this.scrollX) + 0.5;
	      path += "M " + x + " " + (min - this.scrollY) + "L " + x + " " + (maxY - this.scrollY) +" ";
	  }
	  if (i<this.boardSizeY) {
	      var y = Math.floor(i *  this.pointWidth+ this.pointWidth/2 + this.margin - this.scrollY) + 0.5;
	      path += "M " + (min - this.scrollX) + " " + y + "L " + (max - this.scrollX) + " " + y + " ";
	  }
      }
      var el1 = this.jra.path(path).attr({stroke: "#cecfce", "stroke-width": 1});
      el1.toBack();
      this.jra_arr.push(el1);
      path = "";
      for (var i=0; i<maxXY; i+=5) {
	  if (i<this.boardSize) {
	      var x = Math.floor(i *  this.pointWidth+ this.pointWidth/2 + this.margin - this.scrollX) + 0.5;
	      path += "M " + x + " " + (min - this.scrollY) + "L " + x + " " + (maxY - this.scrollY) +" ";
	  }
	  if (i<this.boardSizeY) {
	      var y = Math.floor(i *  this.pointWidth+ this.pointWidth/2 + this.margin - this.scrollY) + 0.5;
	      path += "M " + (min - this.scrollX) + " " + y + "L " + (max - this.scrollX) + " " + y + " ";
	  }
      }
      var el2 = this.jra.path(path).attr({stroke: "#bbbcbb", "stroke-width": 1});
      el2.toBack();
      this.jra_arr.push(el2);
    }
    },
    renderCurrStop0: function() {
    // clear all curr stops
      while (this.jra_arr_cs.length > 0) {
	  var el = this.jra_arr_cs.pop();
	  el.remove();
      }
    },
    renderCurrStop1: function() {
    // paint all curr stops
	//      this.jgcs.paint();
    },

    setCursor: function(cursor) {
        this.domNode.style.cursor = cursor;
    },
    handleHover: function(e) {
        var xy = this.getXY(e);
        this.player.handleBoardHover(xy[0], xy[1], xy[2], xy[3], e);
    },
    handleMouseDown: function(e) {
        var xy = this.getXY(e);
        this.player.handleBoardMouseDown(xy[0], xy[1], xy[2], xy[3], e);
    },
    handleMouseUp: function(e) {
        var xy = this.getXY(e);
        this.player.handleBoardMouseUp(xy[0], xy[1]);
    },
    /**
     *  Gets the board coordinates (0-18) for a mouse event
    **/
    getXY: function(e) {
        var clickXY = eidogo.util.getElClickXY(e, this.domNode);
        
        var m = this.margin;
        var pw = this.pointWidth;
        var ph = this.pointHeight;
        
        var x = Math.round((clickXY[0] - m - (pw / 2)) / pw);
        var y = Math.round((clickXY[1] - m - (ph / 2)) / ph);
    
        return [x, y, clickXY[0], clickXY[1]];
    },
    crop: function(crop) {
        eidogo.util.addClass(this.domContainer, "shrunk");
        this.domGutter.style.overflow = "hidden";
        var width = crop.width * this.pointWidth + (this.margin * 2);
        var height = crop.height * this.pointHeight + (this.margin * 2);
        this.domGutter.style.width = width + "px";
        this.domGutter.style.height = height + "px";
        this.player.dom.player.style.width = width + "px";
        this.domGutter.scrollLeft = crop.left * this.pointWidth;
        this.domGutter.scrollTop = crop.top * this.pointHeight;
    }
}

/**
 * Flash board renderer
**/
eidogo.BoardRendererFlash = function() {
    this.init.apply(this, arguments);
}
eidogo.BoardRendererFlash.prototype = {
    /**
     * @constructor
     * @param {HTMLElement} domContainer Where to put the board
     */
    init: function(domContainer, boardSize, player, crop) {
        if (!domContainer) {
            throw "No DOM container";
            return;
        }
        this.ready = false;
        this.swf = null;
        this.unrendered = [];
        var swfId = domContainer.id + "-board";
        var so = new SWFObject(eidogo.playerPath + "/swf/board.swf", swfId,
            "421", "421", "8", "#665544");
        so.addParam("allowScriptAccess", "sameDomain");
        so.write(domContainer);
        var elapsed = 0;
        var initBoard = function() {
            swf = eidogo.util.byId(swfId);
            if (!swf || !swf.init) {
                if (elapsed > 2000) {            
                    throw "Error initializing board";
                    return;
                }
                setTimeout(arguments.callee.bind(this), 10);
                elapsed += 10;
                return;
            }
            this.swf = swf;
            this.swf.init(player.uniq, boardSize);
            this.ready = true;
        }.bind(this);
        initBoard();
    },
    showRegion: function(bounds) {
    },
    hideRegion: function() {
    },
    clear: function() {
        if (!this.swf) return;
        this.swf.clear();
    },
    renderStone: function(pt, color) {
        if (!this.swf) {
            this.unrendered.push(['stone', pt, color]);
            return;
        }
        for (var i = 0; i < this.unrendered.length; i++) {
            if (this.unrendered[i][0] == "stone") {
                this.swf.renderStone(this.unrendered[i][1], this.unrendered[i][2]);
            }
        }
        this.unrendered = [];
        this.swf.renderStone(pt, color);
    },
    renderMarker: function(pt, type) {
        if (!type) return;
        if (!this.swf) {
            this.unrendered.push(['marker', pt, type]);
            return;
        }
        for (var i = 0; i < this.unrendered.length; i++) {
            if (this.unrendered[i][0] == "marker") {
                this.swf.renderMarker(this.unrendered[i][1], this.unrendered[i][2]);
            }
        }
        this.unrendered = [];
        this.swf.renderMarker(pt, type);
    },
    setCursor: function(cursor) {
    },
    crop: function() {
    }
}

/**
 * @class ASCII board renderer! Kinda broken.
 */
eidogo.BoardRendererAscii = function(domNode, boardSize) {
    this.init(domNode, boardSize);
}
eidogo.BoardRendererAscii.prototype = {
    pointWidth: 2,
    pointHeight: 1,
    margin: 1,
    blankBoard: "+-------------------------------------+\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "|. . . . . . . . . . . . . . . . . . .|\n" +
                "+-------------------------------------+",
    init: function(domNode, boardSize) {
        this.domNode = domNode || null;
        this.boardSize = boardSize || 19;
        this.content = this.blankBoard;
    },
    clear: function() {
        this.content = this.blankBoard;
        this.domNode.innerHTML = "<pre>" + this.content + "</pre>";
    },
    renderStone: function(pt, color) {
        var offset = (this.pointWidth * this.boardSize + this.margin * 2)
            * (pt.y * this.pointHeight + 1)
            + (pt.x * this.pointWidth) + 2;
        this.content = this.content.substring(0, offset-1) + "."
            + this.content.substring(offset);
        if (color != "empty") {
            this.content = this.content.substring(0, offset-1) +
                (color == "white" ? "O" : "#") + this.content.substring(offset);
        }
        this.domNode.innerHTML = "<pre>" + this.content + "</pre>";
    },
    renderMarker: function(pt, type) {
        // I don't think this is possible
    }
}
